/*
 *
 *    OPEN-XCHANGE legal information
 *
 *    All intellectual property rights in the Software are protected by
 *    international copyright laws.
 *
 *
 *    In some countries OX, OX Open-Xchange, open xchange and OXtender
 *    as well as the corresponding Logos OX Open-Xchange and OX are registered
 *    trademarks of the Open-Xchange, Inc. group of companies.
 *    The use of the Logos is not covered by the GNU General Public License.
 *    Instead, you are allowed to use these Logos according to the terms and
 *    conditions of the Creative Commons License, Version 2.5, Attribution,
 *    Non-commercial, ShareAlike, and the interpretation of the term
 *    Non-commercial applicable to the aforementioned license is published
 *    on the web site http://www.open-xchange.com/EN/legal/index.html.
 *
 *    Please make sure that third-party modules and libraries are used
 *    according to their respective licenses.
 *
 *    Any modifications to this package must retain all copyright notices
 *    of the original copyright holder(s) for the original code used.
 *
 *    After any such modifications, the original and derivative code shall remain
 *    under the copyright of the copyright holder(s) and/or original author(s)per
 *    the Attribution and Assignment Agreement that can be located at
 *    http://www.open-xchange.com/EN/developer/. The contributing author shall be
 *    given Attribution for the derivative code and a license granting use.
 *
 *     Copyright (C) 2004-2012 Open-Xchange, Inc.
 *     Mail: info@open-xchange.com
 *
 *
 *     This program is free software; you can redistribute it and/or modify it
 *     under the terms of the GNU General Public License, Version 2 as published
 *     by the Free Software Foundation.
 *
 *     This program is distributed in the hope that it will be useful, but
 *     WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 *     or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
 *     for more details.
 *
 *     You should have received a copy of the GNU General Public License along
 *     with this program; if not, write to the Free Software Foundation, Inc., 59
 *     Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

package com.openexchange.office.json.actions;

import java.io.FilenameFilter;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;

import org.apache.commons.logging.Log;
import org.json.JSONArray;
import org.json.JSONException;
import org.json.JSONObject;

import com.openexchange.ajax.requesthandler.AJAXRequestData;
import com.openexchange.ajax.requesthandler.AJAXRequestResult;
import com.openexchange.documentation.RequestMethod;
import com.openexchange.documentation.annotations.Action;
import com.openexchange.documentation.annotations.Parameter;
import com.openexchange.exception.OXException;
import com.openexchange.file.storage.File;
import com.openexchange.file.storage.composition.IDBasedFileAccess;
import com.openexchange.file.storage.composition.IDBasedFileAccessFactory;
import com.openexchange.folderstorage.FolderService;
import com.openexchange.groupware.ldap.User;
import com.openexchange.groupware.results.TimedResult;
import com.openexchange.office.tools.ConfigurationHelper;
import com.openexchange.office.tools.DocumentType;
import com.openexchange.office.tools.ErrorCode;
import com.openexchange.office.tools.FileHelper;
import com.openexchange.office.tools.LocalFileMappingManager;
import com.openexchange.office.tools.MimeTypeHelper;
import com.openexchange.server.ServiceLookup;
import com.openexchange.tools.iterator.SearchIterator;
import com.openexchange.tools.session.ServerSession;

public class GetTemplateListAction extends DocumentFilterAction {
    private static final Log LOG = com.openexchange.log.Log.loggerFor(GetTemplateListAction.class);

    private static String TYPE_TEXT = "text";
    private static String TYPE_SPREADSHEET = "spreadsheet";
    private static String TYPE_PRESENTATION = "presentation";
    private static boolean DISABLE_ODF_SPREADSHEET = true;
    private static boolean DISABLE_BINARY_MS_FORMATS = true;

    private LocalFileMappingManager m_localFileMapper = null;
    private ExtensionFilter m_extensionFilter = null;

    /**
     * A filter class to filter out language specific folders in the global
     * template folder.
     *
     * {@link LanguageFolderFilter}
     *
     * @author <a href="mailto:carsten.driesner@open-xchange.com">Carsten Driesner</a>
     */
    private static class LanguageFolderFilter implements FilenameFilter {
        private String m_specificLangFolder = null;
        private String m_generalLangFolder = null;

        public LanguageFolderFilter(String specificLangFolder, String generalLangFolder) {
            m_specificLangFolder = specificLangFolder;
            if (!specificLangFolder.equalsIgnoreCase(generalLangFolder)) {
                m_generalLangFolder = generalLangFolder;
            }
        }

        /**
         * Filer method to filter out all folders except the language specific
         * folders.
         *
         * @param file
         *  The file instance to be checked.
         *
         * @param fileName
         *  The file name of the file instance to be checked.
         *
         * @return
         *  TRUE if the file should not be removed from the result set,
         *  otherwise FALSE.
         */
        @Override
        public boolean accept(java.io.File file, String fileName) {
            return (file.isDirectory() &&
                    (fileName.equalsIgnoreCase(m_specificLangFolder)) ||
                    (fileName.equalsIgnoreCase(m_generalLangFolder)));
        }
    }

    /**
     * A filter class to filter out non-templates file according to the file
     * extension.
     *
     * {@link ExtensionFilter}
     *
     * @author <a href="mailto:carsten.driesner@open-xchange.com">Carsten Driesner</a>
     */
    private static class ExtensionFilter implements FilenameFilter {
        private final HashMap<String, Boolean> m_acceptedExtensions = new HashMap<String, Boolean>();

        public ExtensionFilter(String[] acceptedExtensions, String[]...more) {
            if (null != acceptedExtensions) {
	            for (String ext : acceptedExtensions) {
	                m_acceptedExtensions.put(ext, true);
	            }
	            for (String[] strArray : more) {
                    for (String ext : strArray) {
	                    m_acceptedExtensions.put(ext, true);
	                }
                }
            }
        }

        @Override
        public boolean accept(java.io.File file, String fileName) {
            String extension = FileHelper.getExtension(fileName);

            return (file.isFile() && m_acceptedExtensions.containsKey(extension));
        }
    }

    /**
     * {@link GetTemplateListAction}
     *
     * @author <a href="mailto:carsten.driesner@open-xchange.com">Carsten Driesner</a>
     */

    @Action(method = RequestMethod.GET, name = "gettemplatelist", description = "Request to receive a list of possible template files.", parameters = {
        @Parameter(name = "session", description = "A session ID previously obtained from the login module."),
        @Parameter(name = "type", description = "The type of document templates to build up the list from."),
        @Parameter(name = "uid", description = "The unique id of the client application.")
    }, responseDescription = "Response with timestamp: A JSON object containing the list of templates or a certain type.")

    /**
     * Initializes a new {@link GetTemplateListAction}.
     * @param services
     */
    public GetTemplateListAction(ServiceLookup services, LocalFileMappingManager localFileMapper) {
        super(services);
        m_localFileMapper = localFileMapper;
        m_extensionFilter = new ExtensionFilter(MimeTypeHelper.EXCEL_TEMPLATE_EXTENSIONS,
                MimeTypeHelper.WORD_TEMPLATE_EXTENSIONS,
                MimeTypeHelper.POWERPOINT_TEMPLATE_EXTENSIONS,
                MimeTypeHelper.ODF_CALC_TEMPLATE_EXTENSIONS,
                MimeTypeHelper.ODF_IMPRESS_TEMPLATE_EXTENSIONS,
                MimeTypeHelper.ODF_WRITER_TEMPLATE_EXTENSIONS);
    }

    /* (non-Javadoc)
     * @see com.openexchange.ajax.requesthandler.AJAXActionService#perform(com.openexchange.ajax.requesthandler.AJAXRequestData, com.openexchange.tools.session.ServerSession)
     */
    @Override
    public AJAXRequestResult perform(AJAXRequestData request, ServerSession session) {
        final JSONObject jsonObj = new JSONObject();
        final String type = request.getParameter("type").toLowerCase();
        final JSONArray templates = new JSONArray();
        AJAXRequestResult requestResult = null;
        ErrorCode errorCode = ErrorCode.NO_ERROR;

        if ((null != request) && (null != session)) {
            DocumentType searchFilter = DocumentType.NONE;
            IDBasedFileAccess fileAccess = null;

            // Map type to DocumenntType to restrict search filter - other values means all
            if (TYPE_TEXT.equalsIgnoreCase(type)) {
                searchFilter = DocumentType.TEXT;
            } else if (TYPE_SPREADSHEET.equalsIgnoreCase(type)) {
                searchFilter = DocumentType.SPREADSHEET;
            } else if (TYPE_PRESENTATION.equalsIgnoreCase(type)) {
                searchFilter = DocumentType.PRESENTATION;
            }

            try {
                final IDBasedFileAccessFactory fileFactory = m_services.getService(IDBasedFileAccessFactory.class);
                final String folderId = FileHelper.getUserFolderId(m_services.getService(FolderService.class), session);
                final User user = session.getUser();
                final String fallbackTemplatePath = "/opt/open-xchange/templates/documents";
                String userLangCode = "en"; // fallback if no language is set

                fileAccess = fileFactory.createAccess(session);

                if (null != user) {
                    userLangCode = FileHelper.mapUserLanguageToLangCode(user.getPreferredLanguage());
                }

                // scan the user folder and collect available document templates
                errorCode = scanUserFolderForTemplates(fileAccess, templates, folderId, searchFilter);

                // next step: we try to scan the global template path and
                // collect the files there.
                String globalPath = ConfigurationHelper.getStringOfficeConfigurationValue(m_services, session, "//module/templatePath", fallbackTemplatePath);
                if ((null != globalPath) && (globalPath.length() > 0)) {
                    // scan the global folder and collect templates according to the user language
                    errorCode = scanGlobalTemplateFolder(globalPath, templates, searchFilter, userLangCode, m_extensionFilter);
                } else {
                    LOG.error("Error global template path is not set");
                }
            } finally {
                try {
                    jsonObj.put("error", errorCode.getAsJSON());
                    jsonObj.put("templates", templates);
                } catch (JSONException je) {
                    // nothing can be done here
                }
            }
        }

        requestResult = new AJAXRequestResult(jsonObj);

        return requestResult;
    }

    /**
     * Collects the files within the specified folder while checking that no
     * duplicate files are added.
     *
     * @param folder
     *  The folder which should be scanned.
     *
     * @param templates
     *  The JSONArray which will contain the found files.
     *
     * @param duplicates
     *  A hash map which ensures that no duplicate files are collected. The
     *  first occurrence will win and be part of the templates collection.
     *
     * @param extFilter
     *  A filter that filter out all files which don't comply to the
     *  extension filter provided. Can be null so no filtering is done.
     *
     * @return
     *  ErrorCode.NONE on success or an error code if something went wrong.
     */
    private ErrorCode collectTemplateFiles(java.io.File folder, JSONArray templates, HashMap<String, Boolean> duplicates, String type, ExtensionFilter extFilter) {
        java.io.File[] filesAndFolders = folder.listFiles();

        if (null != filesAndFolders) {
            for (java.io.File file : filesAndFolders) {
                if (file.isDirectory()) {
                    // scan sub-folders, too
                    collectTemplateFiles(file, templates, duplicates, type, extFilter);
                } else {
                    // scan the current folder for the files
                    String fileName = file.getName();
                    if (null != fileName) {
                        Boolean exists = duplicates.get(fileName);
                        if (null == exists) {
                            // new entry found, put into map and add it to the
                            // templates
                            duplicates.put(fileName, true);
                            try {
                                // we use the extension to determine the mime type of the file
                                boolean addFile = true;
                                String ext = FileHelper.getExtension(fileName);

                                // use optional file extension filter to filter
                                // out unwanted files.
                                if ((null == extFilter) || ((null != extFilter) && extFilter.accept(file, fileName))) {
	                                String fileMimeType = MimeTypeHelper.getMimeTypeFromTemplateExtension(ext);
	                                long lastModified = file.lastModified();

	                                if (DISABLE_ODF_SPREADSHEET && (MimeTypeHelper.MIMETYPE_VND_OASIS_OPENDOCUMENT_SPREADSHEET_TEMPLATE.equalsIgnoreCase(fileMimeType))) {
	                                    // filter out odf spreadsheet
	                                    addFile = false;
	                                }
	                                if (addFile && DISABLE_BINARY_MS_FORMATS) {
	                                    addFile = !isBinaryMSTemplateMimeType(fileMimeType);
	                                }

	                                if (addFile) {
	                                    putFileData(templates, true, file.getCanonicalPath(), "template:local", file.getName(), file.getName(), "global", type, lastModified, fileMimeType);
	                                }
                                }
                            } catch (JSONException e) {
                                // just ignore
                            } catch (IOException e) {
                                return ErrorCode.GETTEMPLATELIST_GLOBALFOLDER_IO_ERROR;
                            }
                        }
                    }
                }
            }
        }

        return ErrorCode.NO_ERROR;
    }

    /**
     * Scans a set of folders for specific language folders and collects
     * all files within these folders.
     *
     * @param folders
     *  An array of Files that specifies what folders should be scanned for the
     *  language specific folders.
     *
     * @param templates
     *  A JSON array which will collect all template files.
     *
     * @param langCode
     *  The user language code in ISO format.
     *
     * @param extFilter
     *  A filter that filter out all files which don't comply to the
     *  extension filter provided. Can be null so no filtering is done.
     *
     * @return
     *  ErrorCode.NONE on success or an error code if something went wrong.
     */
    private ErrorCode scanForLanguageDocumentTemplates(java.io.File[] folders, JSONArray templates, String langCode, ExtensionFilter extFilter) {
        final String fallbackLangCode = "en";
        ErrorCode errorCode = ErrorCode.NO_ERROR;

        // create the general language code if we have a specific one
        String generalLangCode = langCode;
        if ((null != langCode) && (langCode.indexOf("-") > 0)) {
            generalLangCode = langCode.substring(0, langCode.indexOf("-"));
        }

        // loop through all document folders and search for the concrete
        // language folders within these folders.
        for (java.io.File folder : folders) {
            String folderName = folder.getName();
            String type = null;

            if (TYPE_TEXT.equalsIgnoreCase(folderName)) {
                type = DocumentType.TEXT.toString().toLowerCase();
            } else if (TYPE_SPREADSHEET.equalsIgnoreCase(folderName)) {
                type = DocumentType.SPREADSHEET.toString().toLowerCase();
            } else if (TYPE_PRESENTATION.equalsIgnoreCase(folderName)) {
                type = DocumentType.PRESENTATION.toString().toLowerCase();
            }

            if (null != type) {
                // now we sort the entries for size to search the specific
                // language folder before the general one.
                int numOfTemplates = templates.length();
                HashMap<String, Boolean> duplicates = new HashMap<String, Boolean>();
                java.io.File[] langFoldersFound = folder.listFiles(new LanguageFolderFilter(langCode, generalLangCode));
                if (langFoldersFound.length > 0) {

                    java.util.List<java.io.File> langFolders = Arrays.asList(langFoldersFound);
                    java.util.Collections.sort(langFolders, new java.util.Comparator<java.io.File>() {
                        @Override
                        public int compare(java.io.File f1, java.io.File f2) {
                            return (f1.getName().length() > f2.getName().length()) ? -1 : 1;
                        }
                    });

                    // iterate through the found language folders and collect
                    // the template file.
                    ErrorCode collectErrorCode = ErrorCode.NO_ERROR;
                    for (java.io.File langFolder : langFolders) {
                        collectErrorCode = collectTemplateFiles(langFolder, templates, duplicates, type, extFilter);
                        if (collectErrorCode.getCode() != ErrorCode.NO_ERROR.getCode()) {
                            errorCode = collectErrorCode;
                            break;
                        }
                    }
                }

                // if we didn't find any templates in the specific lang folders
                // and the fallback language code is not langCode or
                // generalLangCode, we try to use it as last possible step
                if ((templates.length() == numOfTemplates) &&
                    (!((fallbackLangCode.equalsIgnoreCase(langCode)) || (fallbackLangCode.equalsIgnoreCase(generalLangCode))))) {
                    langFoldersFound = folder.listFiles(new LanguageFolderFilter(fallbackLangCode, null));
                    if (langFoldersFound.length > 0) {
                        errorCode = collectTemplateFiles(langFoldersFound[0], templates, duplicates, type, extFilter);
                    }
                }
            }
        }

        return errorCode;
    }

    /**
     * Scans the global template folder, trying to find the document specific
     * folders according to the provided searchFilter and the provided user
     * language code.
     *
     * @param path
     *  The path of the "global" templates folder containing the document
     *  type specific sub-folders.
     *
     * @param templates
     *  A JSONArray which will contain the found document templates within
     *  the templates folders.
     *
     * @param searchFilter
     *  A filter containing the type to select the document templates, a
     *  value of DocumentType.NONE means all document types.
     *
     * @param extFilter
     *  A filter that filter out all files which don't comply to the
     *  extension filter provided. Can be null so no filtering is done.
     *
     * @return
     *  ErrorCode.NONE on success or an error code if something went wrong.
     */
    private ErrorCode scanGlobalTemplateFolder(String path, JSONArray templates, DocumentType searchFilter, String userLangCode, ExtensionFilter extFilter) {
        ErrorCode errorCode = ErrorCode.NO_ERROR;
        java.io.File file = new java.io.File(path);

        if (file.exists() && file.isDirectory()) {
            java.io.File[] files = file.listFiles();

            if (DocumentType.NONE == searchFilter) {
                // use all folders, filter out possible files
                ArrayList<java.io.File> folders = new ArrayList<java.io.File>();
                for (java.io.File folder : files) {
                    if (folder.isDirectory()) {
                        folders.add(folder);
                    }
                }

                if (!folders.isEmpty()) {
                    scanForLanguageDocumentTemplates(files, templates, userLangCode, extFilter);
                }
            } else {
                // use just one folder
                String folderToFind = searchFilter.name();
                for (java.io.File folder : files) {
                    if (folder.isDirectory() && (folderToFind.equalsIgnoreCase(folder.getName()))) {
                        java.io.File[] theFolder = new java.io.File[1];
                        theFolder[0] = folder;
                        scanForLanguageDocumentTemplates(theFolder, templates, userLangCode, extFilter);
                        break;
                    }
                }
            }
        }

        return errorCode;
    }

    /**
     * Scans the provided folder for document template files according to the
     * search filter.
     *
     * @param fileAccess
     *  The file access to be used for the method.
     *
     * @param templates
     *  A JSON array that will collect the found document template files.
     *
     * @param folderId
     *  The folder to be searched.
     *
     * @param searchFilter
     *  The search filter which specifies what template documents should be
     *  collected.
     *
     * @return
     *  ErrorCode.NONE on success or an error code if something went wrong.
     */
    private ErrorCode scanUserFolderForTemplates(IDBasedFileAccess fileAccess, JSONArray templates, String folderId, DocumentType searchFilter) {
        ErrorCode errorCode = ErrorCode.NO_ERROR;

        if ((null != folderId) && (null != fileAccess)) {
            try {
                // retrieve all files from the user folder, currently we don't want to
                // search deeper in the folder!
                TimedResult<File> files = fileAccess.getDocuments(folderId);
                if (null != files) {
                    java.util.Date currentDate = new java.util.Date();
                    SearchIterator<File> iter = files.results();

                    while (iter.hasNext()) {
                        File file = iter.next();
                        String ext = FileHelper.getExtension(file.getFileName());
                        String fileMimeType = file.getFileMIMEType();
                        String knownMimeType = MimeTypeHelper.getMimeTypeFromTemplateExtension(ext);
                        DocumentType fileDocType = DocumentType.NONE;
                        java.util.Date lastModifiedDate = file.getLastModified();
                        long lastModified = (null != lastModifiedDate) ? lastModifiedDate.getTime() : currentDate.getTime();

                        // First use the extension to find template documents, second the
                        // mime type of the file. Special handling for binary ms document
                        // template files as they share the same mime type as the normal
                        // documents.
                        if (null != knownMimeType) {
                            fileDocType = MimeTypeHelper.getDocumentTypeFromTemplateMimeType(knownMimeType);
                        } else if (!isBinaryMSTemplateMimeType(fileMimeType)) {
                            fileDocType = MimeTypeHelper.getDocumentTypeFromTemplateMimeType(fileMimeType);
                        }

                        if (isBinaryMSTemplateMimeType(knownMimeType) && DISABLE_BINARY_MS_FORMATS) {
                            fileDocType = null;
                        } else if (DISABLE_ODF_SPREADSHEET) {
                            // filter out odf spreadsheet
                            fileDocType = ((MimeTypeHelper.MIMETYPE_VND_OASIS_OPENDOCUMENT_SPREADSHEET_TEMPLATE.equalsIgnoreCase(knownMimeType)) ||
                                           (MimeTypeHelper.MIMETYPE_VND_OASIS_OPENDOCUMENT_SPREADSHEET_TEMPLATE.equalsIgnoreCase(fileMimeType))) ? null : fileDocType;
                        }

                        if ((null != fileDocType) && (DocumentType.NONE != fileDocType) &&
                            ((DocumentType.NONE == searchFilter) || (searchFilter == fileDocType))) {
                            putFileData(templates, false, file.getId(), folderId, file.getFileName(), file.getTitle(), "user", fileDocType.toString().toLowerCase(), lastModified, file.getFileMIMEType());
                        }
                    }
                }
            } catch (OXException e) {
                errorCode = ErrorCode.GETTEMPLATELIST_UNKNOWN_ERROR;
            } catch (JSONException e) {
                errorCode = ErrorCode.GETTEMPLATELIST_UNKNOWN_ERROR;
            }
        } else {
            return ErrorCode.GETTEMPLATELIST_USERFOLDER_UNKNOWN_ERROR;
        }

        return errorCode;
    }

    /**
     * Put the file data into the provided JSONArray.
     *
     * @param array
     *  The JSONArray which should be used to add a new file data entry to.
     *
     * @param isLocalFile
     *  Must be set to TRUE if the file data specify a local file system
     *  document.
     *
     * @param fileId
     *  The file id of the document template. Must be a OX Drive specific id
     *  or a local file system path (depends on isLocalFile). Must not be null.
     *
     * @param folderId
     *  The folder id of the document template. Must not be null.
     *
     * @param fileName
     *  The file name of the document template. Must not be null.
     *
     * @param title
     *  The title of the document template. Should be set to fileName, if
     *  there is no title. Must not be null.
     *
     * @param source
     *  The source of the document template. Possible values must be "user",
     *  "global", which describe where th document template has been found.
     *  Must not be null.
     *
     * @param type
     *  The document type of the document template. Possible values are:
     *  "text", "spreadsheet" or "presentation". Must not be null.
     *
     * @param lastModified
     *  The last modified date in millseconds since
     *
     * @param mimeType
     *  The mime type of the document template. Can be set to null, if the
     *  mime type is unknown.
     *
     * @throws JSONException
     */
    private void putFileData(JSONArray array, boolean isLocalFile, String fileId, String folderId, String fileName, String title, String source, String type, long lastModified, String mimeType) throws JSONException {
        try {
            JSONObject fileData = new JSONObject();
            fileData.put("folder_id", folderId);
            fileData.put("filename", fileName);
            fileData.put("title", title);
            fileData.put("source", source);
            fileData.put("type", type);
            fileData.put("last_modified", lastModified);
            if (null != mimeType) {
                fileData.put("mimetype", mimeType);
            }

            // Special treatment for local files - we generate a artificial ID
            // which is stored with the path in the LocalFileMappingManager.
            // This helps to protect this code from being used to access other
            // files due to hacked strings.
            if (isLocalFile) {
                String id = m_localFileMapper.getIDFromPath(fileId);
                if (null == id) {
                    id = LocalFileMappingManager.generateIDFromPath(fileId);
                    if (null != id) {
                        m_localFileMapper.associateIdAndPath(id, fileId);
                        // set the artificial ID
                        fileId = id;
                    }
                } else {
                    fileId = id;
                }
            }

            // put entry into the array
            if (null != fileId) {
                fileData.put("id", fileId);
                array.put(fileData);
            }
        } catch (JSONException e) {
            throw e;
        }
    }

    /**
     * Determine if the provided mime type belongs to a binary document
     * template format of MS Office.
     *
     * @param mimeType
     *  The mime type to be checked.
     *
     * @return
     *  TRUE if the mime type belongs to a binary document template format of
     *  MS Office.
     */
    private boolean isBinaryMSTemplateMimeType(String mimeType) {
        return ((MimeTypeHelper.MIMETYPE_MSEXCEL.equalsIgnoreCase(mimeType)) ||
                (MimeTypeHelper.MIMETYPE_MSPOWERPOINT.equalsIgnoreCase(mimeType)) ||
                (MimeTypeHelper.MIMETYPE_MSWORD.equalsIgnoreCase(mimeType)));
    }

}
